/*
 Right, so this project is all about displaying a 2D navigation map for the
 exemplification of binary trees and graphs. This code uses three global arrays 
 of points:
        
 Points: the regular points to display
 HighPoints: the points which will appear highlighted.
 PolyPoints: the points which will form a path (polyline) or polygon.
 
 And two global points to define a rectangle:
 
 RectPoint1: Upper-left corner of the rectangle.
 RectPoint2: Lower-right corner of the rectangle.
 
 That's pretty much it. There's no fancy 3D code in this example. As a matter of
 fact, this project could've been done in a regular windows-based application.
 However, I wanted it to be fancy, so I left it in 3D.
 
 This program implements a thread that expects incoming commands and processes 
 them. The thread blocks until the signal CommandEvent is received (this signal 
 is assumed to be raised by the external Prolog program or however is issuing
 the commands). When a command is issued, Prolog blocks until a signal called 
 CommandProcessedEvent is raised. This signal is raised in this procress when 
 the command has been processed (of course!).
 
 That's pretty much it. The rest are just standard point-display routines in
 OpenGL.
*/
// ******* WIN 32 variables and UI stuff ******* //
#include <windows.h>
#include <windowsx.h>
#include "SharedMem.h"
#include "SharedDefs.h"

#define APP_NAME "Robotic Navigation Uber-Cool Visualizer"

bool MainWindowAvailable = false;
HWND GlobalWindow = NULL;
WNDPROC OldWindowProc = 0;
float DifMouseXAngle, DifMouseYAngle, DifMouseZTrans, DifMouseXTrans, DifMouseYTrans;
float LastDifMouseXAngle, LastDifMouseYAngle, LastDifMouseZTrans, LastDifMouseXTrans, LastDifMouseYTrans;
bool DoProc = false;
int LastMouseX, LastMouseY, CurMouseX, CurMouseY;

// ****** The real program begins here ****** //

#include <iostream>
#include <stdlib.h>
#include <math.h>
#include <GL/glut.h>
#include <fstream>

using namespace std;

// The ordinary points which make up the map
Point2D Points[MAX_POINTS];
int nPoints = 0;

//Point2D Points2[MAX_POINTS];
//int nPoints2 = 0;

// The highlighted points...
Point2D HighPoints[MAX_POINTS];
int nHighPoints = 0;

// The rectangle to display
Point2D RectPoint1, RectPoint2;
bool DisplayRectangle = false;

// The polygon to display... Ahoy!
Point2D PolyPoints[MAX_POINTS];
int nPolyPoints = 0;
bool PolyLine = false;

ostream &operator<<(ostream &out, const Point2D &P) {
  out << "(" << P.x << ", " << P.y << ")";
  return out;
}

// The pointer to the shared prolog data...
PrologData *SharedPrologData = NULL; 

// The shared memory object which takes care of all the low-level details.
SharedMem *sharedmem = NULL;

void GenPointsTest() {
  // This method is here just to test that stuff is being displayed OK.
  Points[0].x = 0;
  Points[0].y = -20;
  
  Points[1].x = 20;
  Points[1].y = 0;
  
  Points[2].x = 0;
  Points[2].y = 20;

  Points[3].x = -20;
  Points[3].y = 0;
  
  nPoints = 4;
}

// ************************************************************************** //
// ************************************************************************** //
// ************************************************************************** //

// In this section we will declare the thread that receives the commands
// (from Prolog) and processes them.

bool RunThread = true;
bool WaitForEvent = true;
HANDLE ThreadHandle = NULL, CommandEvent = NULL, CommandProcessedEvent = NULL;

DWORD WINAPI ThreadFunction(LPVOID lpParam) { 
    while (RunThread) {
      if (WaitForEvent) {
        if (WaitForSingleObject(CommandEvent, 0) == WAIT_OBJECT_0) {
           cerr << "Command received!" << endl;
           // The CommandEvent will be an AutoReset event, so we don't need
           // to reset it manually here.

           // Now, process the contents of the shared object and see what
           // Prolog wants us to do... We assume SharedPrologData already
           // points to the shared memory buffer
           switch (SharedPrologData->Command) {
             case CMD_NOTHING : 
                  cerr << "Command: CMD_NOTHING" << endl;
                  cerr << "What do you mean? An African or an European swallow?" << endl;
                  break;

             case CMD_SET_DATA : 
                  cerr << "Command: CMD_SET_DATA" << endl;
                  // Copy over the contents of the buffer. Formally, this copy-over
                  // *should* be synch'ed with the thread that displays the data...
                  // ... but whatever!
                  for (int i = 0; i < SharedPrologData->nPoints && i < MAX_POINTS; i++) {
                    //cerr << "i: " << i << ", (" << SharedPrologData->Points[i].x << ", " << SharedPrologData->Points[i].y << ")" << endl;
                    Points[i] = SharedPrologData->Points[i];
                  }
                  nPoints = SharedPrologData->nPoints;
                  break;
                  
             case CMD_SET_HIGHLIGHTED_DATA : 
                  cerr << "Command: CMD_SET_HIGHLIGHTED_DATA" << endl;
                  // No synchronization here either... Shame on me! Bad cow... 
                  // Bad Cow!
                  for (int i = 0; i < SharedPrologData->nPoints && i < MAX_POINTS; i++) {
                    //cerr << "i: " << i << ", (" << SharedPrologData->Points[i].x << ", " << SharedPrologData->Points[i].y << ")" << endl;
                    HighPoints[i] = SharedPrologData->Points[i];
                  }
                  nHighPoints = SharedPrologData->nPoints;
                  break;

             case CMD_SET_POLYLINE :
             case CMD_SET_POLYGON :              
                  PolyLine = SharedPrologData->Command == CMD_SET_POLYLINE;
                  if (PolyLine)
                    cerr << "Command: CMD_SET_POLYLINE" << endl;
                  else
                    cerr << "Command: CMD_SET_POLYGON" << endl;
                  
                  // No synchronization here either... Shame on me! Bad cow... Bad Cow!
                  for (int i = 0; i < SharedPrologData->nPoints && i < MAX_POINTS; i++)
                    PolyPoints[i] = SharedPrologData->Points[i];                  
                  nPolyPoints = SharedPrologData->nPoints;
                  break;
                  
             case CMD_SET_RECTANGLE :
                  cerr << "Command: CMD_SET_RECTANGLE" << endl;
                  DisplayRectangle = SharedPrologData->nPoints > 0;
                  RectPoint1 = SharedPrologData->Points[0];
                  RectPoint2 = SharedPrologData->Points[1];
                  break;
           }

           // Signal the CommandProcessedEvent so that Prolog may continue
           // executing happily away
           SetEvent(CommandProcessedEvent);
        }
      }
    }
    return 0; 
}

void SetupThread() {
  CommandEvent = CreateEvent(
     NULL,                 // Default security attributes.
     FALSE,                // Manual reset?
     FALSE,                // Intial state.
     "OpenGLCommandEvent"  // Event's name. In this case, this is not optional. This name will be used by Prolog to refer to the same object.
  );
  if (CommandEvent == NULL) {
    cerr << "Error! CommandEvent could not be created." << endl;
    return;
  }
  
  CommandProcessedEvent = CreateEvent(
     NULL,                 // Default security attributes.
     FALSE,                // Manual reset?
     FALSE,                // Intial state.
     "OpenGLCommandProcessedEvent"  // Event's name. In this case, this is not optional. This name will be used by Prolog to refer to the same object.
  );
  if (CommandProcessedEvent == NULL) {
    cerr << "Error! CommandProcessedEvent could not be created." << endl;
    return;
  }

  ThreadHandle = CreateThread( 
        NULL,                        // Default security attrs.
        0,                           // Default stack size.
        ThreadFunction,              // Thread function to execute.
        NULL,                        // Argument for the thread.
        0,                           // Default creation flags.
        NULL);                       // Do we want the thread's id? 

  if (ThreadHandle == NULL) {
    cerr << "Error! Command thread could not be created." << endl;
    return;
  }

}

void CloseThread() {
    WaitForEvent = false;
    RunThread = false;
    
    // We signal the CommandProcessedEvent just in case Prolog is waiting for us...
    SetEvent(CommandProcessedEvent);

    // We wait for the thread to actually end...
    cerr << "Waiting for thread to finish..." << endl;
    WaitForSingleObject(ThreadHandle, INFINITE); 
    // ... and then we close the handles.
    CloseHandle(ThreadHandle);
    CloseHandle(CommandEvent);
    CloseHandle(CommandProcessedEvent);
}

// ************************************************************************** //
// ************************************************************************** //
// ************************************************************************** //

// Rotates a vector [x,y,z] around the UNITARY vector [ax,ay,az]
// The equations are based on the Rotor (Quaternion) transformation style (Geometric Algebra)
// I did this because I was too lazy to implement the matrix multiplication which
// is the traditional method to perform the rotation... In retrospective, I have
// no idea why I did this, I spent more time factoring out the equations... Go figure.
void rotateVector(float &x, float &y, float &z,
				  float ax, float ay, float az, float halfsin, float halfcos) {
  float wx = halfcos*x - halfsin*(z*ay - y*az);
  float wy = halfcos*y - halfsin*(x*az - z*ax);
  float wz = halfcos*z - halfsin*(y*ax - x*ay);
  float w = halfsin*(x*ax + y*ay + z*az);

  x = halfcos * wx + halfsin * (ax*w - ay*wz + az*wy);
  y = halfcos * wy + halfsin * (ay*w - az*wx + ax*wz);
  z = halfcos * wz + halfsin * (az*w - ax*wy + ay*wx);
}

// ********* OPEN GL variables ********* //

// General rotation and translation of the scene
float AngX, AngY, AngZ, ScaleFactor;
float TransX, TransY, TransZ;
GLUquadricObj *GeneralQuadric;

// Window data
int WinWidth, WinHeight, MainWinId;

// Lights.
float whiteLight[] = { 1.0f, 1.0f, 1.0f, 1.0f };
float lightPos[2][4] = { 
	{ 0, 0, 1, 0 },
	{ 0, -1, 0, 0 }
};

void SetInitialPos() {
  TransX = 0; TransY = 0; TransZ = -500;
  AngX = 180; AngY = 0; AngZ = 0; ScaleFactor = 1;
  
  DifMouseXAngle = DifMouseYAngle = 0;
  DifMouseZTrans = DifMouseXTrans = DifMouseYTrans = 0;
  DoProc = false;
}

void Setup3D() {
  GeneralQuadric = gluNewQuadric();
  SetInitialPos();
}

void EndProgram() {
  gluDeleteQuadric(GeneralQuadric);
  glutDestroyWindow( MainWinId );
  //system("PAUSE");
  exit(0);
}

void ChangeSize(GLsizei w, GLsizei h) {
	// Prevent division by 0
	if (h == 0) h = 1;

	// Store window's width and height.
	WinWidth = w; WinHeight = h;
	// Set the viewport.
	glViewport(0,0, w,h);

	// Reset transform
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();

	float Aspect = (float )w/(float )h;
	gluPerspective(30.0f, Aspect, 1.0f, 20000.0f);	
	glMatrixMode(GL_MODELVIEW);
}

void DrawString(int x, int y, char *msg) {
	register int i = 0;
	glPushMatrix();	
	glRasterPos3f(x, y, -300);
    while (msg[i])
	  // GLUT_BITMAP_HELVETICA_10
      // GLUT_BITMAP_HELVETICA_12
      // GLUT_BITMAP_HELVETICA_18
      glutBitmapCharacter(GLUT_BITMAP_HELVETICA_18, msg[i++]);	  
	glPopMatrix();
}

void DrawThings() {
    glLineWidth(1);
	glColor3f(1,1,0);
	glBegin(GL_LINES);
		glVertex3f(0,0,0);
		glVertex3f(10,0,0);

		glVertex3f(0,0,0);
		glVertex3f(0,10,0);

		glVertex3f(0,0,0);
		glVertex3f(0,0,10);
	glEnd();
	
	glColor3f(0,0,0);
	glPointSize(3);
	glBegin(GL_POINTS);
      for (int i = 0; i < nPoints; i++)
          glVertex3f(Points[i].x, Points[i].y, 0);
	glEnd();
	
	glColor3f(1,0,0);
	glPointSize(5);
	glBegin(GL_POINTS);
      for (int i = 0; i < nHighPoints; i++)
          glVertex3f(HighPoints[i].x, HighPoints[i].y, 0);
	glEnd();
	
	if (DisplayRectangle) {
      glBegin(GL_LINES);
        glVertex3f(RectPoint1.x, RectPoint1.y, 0);
        glVertex3f(RectPoint2.x, RectPoint1.y, 0);
        
        glVertex3f(RectPoint2.x, RectPoint1.y, 0);
        glVertex3f(RectPoint2.x, RectPoint2.y, 0);
        
        glVertex3f(RectPoint2.x, RectPoint2.y, 0);
        glVertex3f(RectPoint1.x, RectPoint2.y, 0);
        
        glVertex3f(RectPoint1.x, RectPoint2.y, 0);
        glVertex3f(RectPoint1.x, RectPoint1.y, 0);
      glEnd();
    }
    
    if (nPolyPoints > 0) {
       glColor3f(0,0,0.7);
       glLineWidth(2);
       glBegin(GL_LINES);
         for (int i = 0; i < nPolyPoints - 1; i++) {
           glVertex3f(PolyPoints[i].x, PolyPoints[i].y, 0);
           glVertex3f(PolyPoints[i+1].x, PolyPoints[i+1].y, 0);
         }
         if (!PolyLine) {
           glVertex3f(PolyPoints[nPolyPoints-1].x, PolyPoints[nPolyPoints-1].y, 0);
           glVertex3f(PolyPoints[0].x, PolyPoints[0].y, 0);
         }
	   glEnd();
    }
}

void RenderScene() {
    glClearColor(1,1,1, 1);
	glClear(GL_COLOR_BUFFER_BIT);
	glEnable(GL_COLOR_MATERIAL);
	//glEnable(GL_LIGHTING);
	glEnable(GL_NORMALIZE);
	glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE);
	glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, 1);
	// Load identity matrix.
	glLoadIdentity();
	
	// Turn on the lights...
	// Enable light 0
	//glLightfv(GL_LIGHT0, GL_AMBIENT_AND_DIFFUSE, whiteLight);
	//glLightfv(GL_LIGHT0, GL_POSITION, lightPos[0]);
	//glEnable(GL_LIGHT0);
		
	// Local transformation	
	glTranslatef(TransX, TransY, TransZ);	
	glRotatef(AngX, 1, 0, 0);
	glRotatef(AngY, 0, 1, 0);
	glRotatef(AngZ, 0, 0, 1);	
	glScalef(ScaleFactor, ScaleFactor, ScaleFactor);
	
	// Enable light 1
	//glLightfv(GL_LIGHT1, GL_AMBIENT_AND_DIFFUSE, whiteLight);
	//glLightfv(GL_LIGHT1, GL_POSITION, lightPos[1]);	
	//glEnable(GL_LIGHT1);

	// Add a final transformation to reflect the changes performed by the mouse
	glTranslatef(DifMouseXTrans, DifMouseYTrans, DifMouseZTrans);
	glRotatef(DifMouseXAngle, 0,-1,0);
	glRotatef(DifMouseYAngle, 1,0,0);

	DrawThings();
	// Now, display the product of the current render
	glutSwapBuffers();
}

void KeyEvent(unsigned char key, int x, int y) {
	switch (key) {
		// Pressing ESC terminates the program
		case 27 : EndProgram(); break;
		// Rotations.
		case 'a' : case 'A' : AngY += 5; if (AngY > 360) AngY -= 360; break; 
		case 's' : case 'S' : AngY -= 5; if (AngY < 0) AngY += 360; break; 

		case 'x' : case 'X' : AngX += 5; if (AngX > 360) AngX -= 360; break; 
		case 'z' : case 'Z' : AngX -= 5; if (AngX < 0) AngX += 360; break; 

		case 'q' : case 'Q' : AngZ += 5; if (AngZ > 360) AngZ -= 360; break; 
		case 'w' : case 'W' : AngZ -= 5; if (AngZ < 0) AngZ += 360; break; 
		// Translations
		case 'd' : case 'D' : TransX += 50; break; 
		case 'f' : case 'F' : TransX -= 50; break; 

		case 'c' : case 'C' : TransY += 50; break; 
		case 'v' : case 'V' : TransY -= 50; break; 

		case 'e' : case 'E' : TransZ += 500; break; 
		case 'r' : case 'R' : TransZ -= 500; break; 
		// Scale
		case 't' : case 'T' : if (ScaleFactor < 64) ScaleFactor *= 2.0f; break; 
		case 'y' : case 'Y' : if (ScaleFactor > 0.0001) ScaleFactor /= 2.0f; break; 
		// Return to the original position
		case ' ' : SetInitialPos(); break;
	}
}

/*
  This function overrides the hidden window message dispatcher set-up by GLUT. 
  It is mainly used to listen and respond to mouse events:
  Left clicking and dragging around rotates the 3D view
  Right clicking and dragging up and down zooms
  Left and Right clicking and dragging, moves the object around
*/
LRESULT CALLBACK WindowProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam) {  
  switch(uMsg) {
	case WM_LBUTTONDOWN : 
	case WM_RBUTTONDOWN :
		SetCapture(hwnd); // Listen to mouse events even when the mouse has left the window area
		DoProc = true;
		if (wParam & MK_LBUTTON && !(wParam & MK_RBUTTON)) {
			LastMouseX = GET_X_LPARAM(lParam);
			LastMouseY = GET_Y_LPARAM(lParam);
			LastDifMouseXAngle = DifMouseXAngle;
			LastDifMouseYAngle = DifMouseYAngle;
			return 0;
		} else if (!(wParam & MK_LBUTTON) && wParam & MK_RBUTTON) {
			LastMouseY = GET_Y_LPARAM(lParam);			
			LastDifMouseZTrans = DifMouseZTrans;
			return 0;
		} else if (wParam & MK_LBUTTON && wParam & MK_RBUTTON) {
			LastMouseX = GET_X_LPARAM(lParam);
			LastMouseY = GET_Y_LPARAM(lParam);
			LastDifMouseXTrans = DifMouseXTrans;
			LastDifMouseYTrans = DifMouseYTrans;
			return 0;
		}		

	case WM_MOUSEMOVE : 
		if (wParam & MK_LBUTTON && !(wParam & MK_RBUTTON) && DoProc) {
			CurMouseX = GET_X_LPARAM(lParam);
			CurMouseY = GET_Y_LPARAM(lParam);
			int DifX = CurMouseX - LastMouseX;
			int DifY = CurMouseY - LastMouseY;
			DifMouseXAngle = LastDifMouseXAngle + DifX;
			DifMouseYAngle = LastDifMouseYAngle + DifY;
			DifMouseXAngle = (int )DifMouseXAngle % 360;
			DifMouseYAngle = (int )DifMouseYAngle % 360;
			return 0;
		} else if (wParam & MK_LBUTTON && wParam & MK_RBUTTON && DoProc) {
			CurMouseX = GET_X_LPARAM(lParam);
			CurMouseY = GET_Y_LPARAM(lParam);
			int DifX = CurMouseX - LastMouseX;
			int DifY = CurMouseY - LastMouseY;
			DifMouseXTrans = LastDifMouseXTrans + (float )(DifX);
			DifMouseYTrans = LastDifMouseYTrans + (float )(DifY);
			return 0;
		} else if (!(wParam & MK_LBUTTON) && wParam & MK_RBUTTON && DoProc) {
			CurMouseY = GET_Y_LPARAM(lParam);
			int DifY = CurMouseY - LastMouseY;
			DifMouseZTrans = LastDifMouseZTrans + (float )(DifY);
			return 0;
		}
		break; // Return default processing if left mouse button is not down

	case WM_LBUTTONUP :
	case WM_RBUTTONUP :
		ReleaseCapture();
		DoProc = false;
		return 0;
  }

  // Call the default WindowProc handler
  return CallWindowProc(OldWindowProc, hwnd, uMsg, wParam, lParam);  
}

// This timer is used to snatch the default window message dispatcher and replace it with
// our own custom windowproc (OMGHAX!!!). I suppose there are better and easier ways
// to do this; but... whatever. After the message dispatcher has been modified, the
// timer is de-activated. If, for some reason, you wish to continue calling the timer
// the lines glutTimerFunc(1000,...) should be uncommented. The number 1000 is in 
// milliseconds, by the way.
void Timer(int value) {	
	if (!MainWindowAvailable) {
		GlobalWindow = FindWindow(NULL, APP_NAME);
		if (GlobalWindow) {
			// Replace this window's WindowProc
			OldWindowProc = (WNDPROC )SetWindowLong(GlobalWindow, GWL_WNDPROC, (LONG )WindowProc);
			if (OldWindowProc) {
				cerr << "Original WindowProc modified!!!" << endl;
			}
		}
		MainWindowAvailable = true;
		//glutTimerFunc(1000, Timer, value);
		return;
	}
	//glutTimerFunc(1000, Timer, value);
}

// This function is called continuously. It can be used to perform animations, but now we're
// simply using it to redisplay the whole scene as fast as the machine can do it.
void IdleFunc() {	
    glutPostRedisplay();	
}

int main(int argc, char *argv[]) {
    glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
    glutInitWindowSize(640,480);  
    MainWinId = glutCreateWindow(APP_NAME);
    Setup3D();
    GenPointsTest();
    //ReadPoints();

    // Set-up the shared memory buffer...
    sharedmem = new SharedMem(SHAREDMEM_NAME, sizeof(PrologData));
    // Get the pointer to the shared memory buffer
    SharedPrologData = (PrologData *)sharedmem->GetMemory();
    // Start listening for remote Prolog commands...
    SetupThread();
    
    glutReshapeFunc(ChangeSize);
    glutDisplayFunc(RenderScene);
    glutKeyboardFunc(KeyEvent);	
    glutTimerFunc(1, Timer, 0);
    glutIdleFunc(IdleFunc);
    
    glutMainLoop();
    delete sharedmem;
    return 0;
}
